package dev.pfaff.altargui.async;

import dev.pfaff.altargui.util.TaskException;
import org.jspecify.annotations.Nullable;

import java.lang.invoke.MethodHandles;
import java.lang.invoke.VarHandle;
import java.util.concurrent.Callable;
import java.util.concurrent.Executor;
import java.util.concurrent.Phaser;

import static java.util.Objects.requireNonNull;

/**
 * This executor itself is stateless, although the delegate executor may not be.
 */
public final class BorrowedThreadTaskExecutor implements TaskExecutor {
	private final Thread thread;
	private final Executor executor;

	public BorrowedThreadTaskExecutor(Thread thread, Executor executor) {
		this.thread = requireNonNull(thread);
		this.executor = requireNonNull(executor);
	}

	@Override
	public <V> Task<V> submit(Callable<V> callable) {
		var task = new TaskImpl<V>(this.thread);
		this.executor.execute(() -> {
			if (!task.begin()) {
				task.complete(null, new TaskException.Cancelled("Cancelled before beginning"));
				return;
			}
			@Nullable V value;
			TaskException exception;
			try {
				value = callable.call();
				exception = null;
			} catch (InterruptedException e) {
				value = null;
				exception = new TaskException.Cancelled("Borrowed thread interrupted, task cancelled", e);
			} catch (Throwable e) {
				value = null;
				exception = new TaskException.Failed(e.getMessage(), e);
			}
			task.complete(value, exception);
		});
		return task;
	}

	private static final class TaskImpl<V> implements Task.Cancellable<V> {
		private static final byte STATE_PENDING = 0;
		private static final byte STATE_CANCELLING = 1;
		private static final byte STATE_BEGUN = 2;

		private static final VarHandle VH_state;

		static {
			try {
				VH_state = MethodHandles.lookup()
						.findVarHandle(TaskImpl.class, "state", byte.class)
						.withInvokeExactBehavior();
			} catch (NoSuchFieldException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}
		}

		private final Thread thread;
		private final Phaser phaser = new Phaser(1);
		@Nullable
		private V value;
		@Nullable
		private TaskException exception;
		private volatile byte state = STATE_PENDING;

		private TaskImpl(Thread thread) {
			this.thread = thread;
		}

		/**
		 * @return true if the task should be executed, false if it has been cancelled.
		 */
		private boolean begin() {
			assert Thread.currentThread() == this.thread : "Wrong thread";
			assert this.phaser.getPhase() == 0 : "Already completed";
			assert this.state != STATE_BEGUN : "Already begun";
			return VH_state.compareAndSet(this, STATE_PENDING, STATE_BEGUN);
		}

		private void complete(@Nullable V value, @Nullable TaskException exception) {
			assert Thread.currentThread() == this.thread : "Wrong thread";
			assert this.phaser.getPhase() == 0 : "Already completed";
			assert this.state > STATE_PENDING : "Not begun or cancelling";
			// does not need to be atomic thanks to atomic usage of flags.
			this.value = value;
			this.exception = exception;
			this.phaser.arriveAndAwaitAdvance();
			this.phaser.forceTermination();
			// clear cancellation
			var _ = Thread.interrupted();
		}

		@Override
		public V join() throws TaskException, InterruptedException {
			if (this.phaser.register() == 0) {
				this.phaser.awaitAdvanceInterruptibly(this.phaser.arriveAndDeregister());
			}
			if (this.exception != null) {
				throw this.exception;
			}
			//noinspection DataFlowIssue
			return this.value;
		}

		@Override
		public V joinUninterruptibly() throws TaskException {
			if (this.phaser.register() == 0) {
				this.phaser.awaitAdvance(this.phaser.arriveAndDeregister());
			}
			if (this.exception != null) {
				throw this.exception;
			}
			//noinspection DataFlowIssue
			return this.value;
		}

		@Override
		public void cancel() {
			if (!VH_state.compareAndSet(this, STATE_PENDING, STATE_CANCELLING)) {
				if (this.phaser.register() == 0) {
					this.thread.interrupt();
				}
			}
		}
	}
}
