package dev.pfaff.altargui.async;

import dev.pfaff.altargui.function.FallibleRunnable;
import dev.pfaff.altargui.util.CancellationException;
import org.jspecify.annotations.Nullable;

import java.util.concurrent.Callable;
import java.util.concurrent.CompletionException;

import static dev.pfaff.altargui.internal.Config.TRACE_THREAD_FUTURE_SPAWN_POINT;

public final class VirtualThreadTask<V> implements Task.Cancellable<V> {
	private final Thread thread;
	@Nullable
	private final SpawnTrace spawnTrace;

	private V value;
	@Nullable
	private Throwable throwable;

	private VirtualThreadTask(FallibleRunnable runnable) {
		this.thread = Thread.startVirtualThread(() -> {
			try {
				runnable.run();
			} catch (Throwable e) {
				this.throwable = e;
			}
		});
		this.spawnTrace = TRACE_THREAD_FUTURE_SPAWN_POINT ? new SpawnTrace() : null;
	}

	private VirtualThreadTask(Callable<V> callable) {
		this.thread = Thread.startVirtualThread(() -> {
			try {
				this.value = callable.call();
			} catch (Throwable e) {
				this.throwable = e;
			}
		});
		this.spawnTrace = TRACE_THREAD_FUTURE_SPAWN_POINT ? new SpawnTrace() : null;
	}

	public static VirtualThreadTask<Void> spawnVirtualTask(FallibleRunnable runnable) {
		return new VirtualThreadTask<>(runnable);
	}

	public static <V> VirtualThreadTask<V> spawnVirtualTask(Callable<V> callable) {
		return new VirtualThreadTask<>(callable);
	}

	public V join() throws CompletionException, CancellationException, InterruptedException {
		this.thread.join();
		return postJoin();
	}

	@Override
	public V joinUninterruptibly() throws CompletionException, CancellationException {
		while (true) {
			try {
				this.thread.join();
				break;
			} catch (InterruptedException e) {
				// try again
			}
		}
		return postJoin();
	}

	private V postJoin() {
		if (this.throwable != null) {
			RuntimeException e;
			if (this.throwable instanceof CancellationException) {
				e = new CancellationException("Thread task cancelled", this.throwable);
			} else {
				e = new CompletionException("Thread task completed with exception", this.throwable);
			}
			if (TRACE_THREAD_FUTURE_SPAWN_POINT) {
				assert this.spawnTrace != null;
				e.addSuppressed(this.spawnTrace);
			}
			throw e;
		}
		return this.value;
	}

	@Override
	public void cancel() {
		this.thread.interrupt();
	}

	public boolean isAlive() {
		return this.thread.isAlive();
	}

	public Thread thread() {
		return this.thread;
	}

	private static final class SpawnTrace extends Throwable {
	}
}
