package dev.pfaff.altargui.pathfinder.fontkit;

import dev.pfaff.altargui.libimage.ColorType;
import dev.pfaff.altargui.libimage.ImageData;
import dev.pfaff.altargui.text.Font;
import dev.pfaff.altargui.text.GlyphRef;
import dev.pfaff.altargui.texture.LoadableTexture;
import dev.pfaff.altargui.util.GLThread;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.util.Objects;
import java.util.function.IntFunction;

import static dev.pfaff.altargui.pathfinder.fontkit.FKGlyphLoadableTexture.RASTERIZATION_OPTIONS;
import static dev.pfaff.altargui.pathfinder.fontkit.FKGlyphLoadableTexture.unpremultiplyAlpha;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public final class FKFont implements Font {
	private static final int GLYPH_SIZE = 32;
	private static final int PAGE_DIM_GLYPHS = 16;
	private static final int PAGE_DIM_SIZE = PAGE_DIM_GLYPHS * GLYPH_SIZE;
	private static final int PAGE_GLYPHS = PAGE_DIM_GLYPHS * PAGE_DIM_GLYPHS;
	private static final int PAGE_SIZE = PAGE_DIM_SIZE * PAGE_DIM_SIZE;

	private final FKFontRef fkFontRef;
	private final Int2ObjectOpenHashMap<Page> pages = new Int2ObjectOpenHashMap<>();

	public FKFont(FKFontRef font) {
		this.fkFontRef = Objects.requireNonNull(font);
	}

	@Override
	public GlyphRef getGlyph(int codepoint) {
		// TODO: async rasterization
		int pageIdx = codepoint / PAGE_GLYPHS;
		var page = this.pages.computeIfAbsent(pageIdx, (IntFunction<Page>) this::rasterizePage);
		var glyphIdx = codepoint % PAGE_GLYPHS;
		return new GlyphRef(page, (byte) glyphIdx, 0f, page.widths[glyphIdx]);
	}

	private Page rasterizePage(int pageIdx) {
		int[] ids = new int[PAGE_GLYPHS];
		float[] widths = new float[PAGE_GLYPHS];

		int codepointBase = pageIdx * PAGE_GLYPHS;
		for (int i = 0; i < PAGE_GLYPHS; i++) {
			int codepoint = codepointBase + i;

			@SuppressWarnings("DataFlowIssue")
			var idOrNull = GLThread.executor().execute(() -> this.fkFontRef.glyphForCodepoint(codepoint));
			//noinspection ConstantValue
			if (idOrNull == null) continue;
			int id = idOrNull;
			ids[i] = id;
			var rasterBounds = GLThread.executor().execute(() -> this.fkFontRef.rasterBounds(id, GLYPH_SIZE, RASTERIZATION_OPTIONS));
			widths[i] = rasterBounds.width() / (float) GLYPH_SIZE;
		}
		return new Page(this.fkFontRef, ids, widths);
	}

	private static final class Page implements LoadableTexture {
		private final FKFontRef fkFontRef;
		private final int[] ids;
		private final float[] widths;

		private Page(FKFontRef fkFontRef, int[] ids, float[] widths) {
			this.fkFontRef = fkFontRef;
			this.ids = ids;
			this.widths = widths;
		}

		@Override
		public ImageData load() throws IOException {
			var fkFontRef = this.fkFontRef;
			var ids = this.ids;

			var buf = Arena.ofAuto().allocate(JAVA_INT, PAGE_SIZE);
			for (int i = 0; i < PAGE_GLYPHS; i++) {
				int x = i % PAGE_DIM_GLYPHS;
				int y = i / PAGE_DIM_GLYPHS;
				int id = ids[i];

				var translationX = x * GLYPH_SIZE;
//		var translationY = rasterBounds.originY() * this.factors.originY + rasterBounds.height() * this.factors.height;
				var translationY = y * GLYPH_SIZE + GLYPH_SIZE;
				GLThread.executor().execute(() -> fkFontRef.rasterizeCodepoint(buf,
						PAGE_DIM_SIZE,
						PAGE_DIM_SIZE,
						FKCanvasFormat.Rgba32,
						id,
						GLYPH_SIZE,
						translationX,
						translationY,
						RASTERIZATION_OPTIONS));
			}
			unpremultiplyAlpha(buf);
			return new ImageData(PAGE_DIM_SIZE,
					PAGE_DIM_SIZE,
					ColorType.Rgba8,
					buf.asReadOnly());
		}
	}
}
